// Copyright (C) 2015 ChaosForge Ltd
// http://chaosforge.org/
//
// This file is part of NV Libraries.
// For conditions of distribution and use, see copyright notice in nv.hh

/**
* @file memory.hh
* @author Kornel Kisielewicz
* @brief memory utilities
*/

// TODO: implement const and mutable buffer using the union trick

#ifndef NV_CORE_MEMORY_HH
#define NV_CORE_MEMORY_HH

#include <nv/core/common.hh>
#include <nv/stl/traits/properties.hh>
#include <nv/stl/traits/alignment.hh>
#include <nv/stl/utility.hh>
#include <nv/stl/iterator.hh>
#include <nv/stl/capi.hh>

namespace nv
{
	namespace detail
	{
		template< typename T >
		struct addressof_helper
		{
			T & value;
			constexpr addressof_helper( T & v ) : value( v ) {}
			constexpr operator T& () const { return value; }
		private:
			addressof_helper & operator=( const addressof_helper & );
		};

		template< typename T >
		struct addressof_impl
		{
			static constexpr T * f( T & v, long )
			{
				return reinterpret_cast<T*>(
					&const_cast<char&>( reinterpret_cast<const volatile char &>( v ) ) );
			}
			static constexpr T * f( T * v, int ) { return v; }
		};
	}

	template< typename T >
	T * addressof( T & v )
	{
		return detail::addressof_impl<T>::f( detail::addressof_helper<T>( v ), 0 );
	}

	namespace mem_flags
	{
		static constexpr uint16 is_const  = 0x0001;
		static constexpr uint16 is_static = 0x0002;
		static constexpr uint16 read_only = 0x0004;
		static constexpr uint16 temporary = 0x0008;
	}

	template< typename T >
	class storage_view
	{
	public:
		typedef T      value_type;
		typedef size_t size_type;
		static constexpr bool is_static   = false;
		static constexpr bool is_fixed    = false;
		static constexpr bool is_const    = false;

		constexpr storage_view() 
			: m_data( nullptr ), m_size( 0 ) {}
		constexpr storage_view( value_type* a_data, size_type a_size )
			: m_data( a_data ), m_size( a_size ) {}

		void assign( value_type* a_data, size_type a_size )
		{
			m_data = a_data;
			m_size = a_size;
		}

		constexpr size_t size() const { return m_size; }
		constexpr bool empty() const { return m_size != 0; }
		constexpr const value_type* data() const { return m_data; }
		inline    value_type* data() { return m_data; }
		constexpr size_type   raw_size() const { return sizeof( value_type ) * m_size; }
		constexpr const char* raw_data() const { return reinterpret_cast<const char*>( m_data ); }
		inline    char* raw_data() { return reinterpret_cast<char*>( m_data ); }
	protected:
		value_type* m_data;
		size_type   m_size;
	};

	template< typename T >
	class const_storage_view
	{
	public:
		typedef T        value_type;
		typedef size_t   size_type;
		static constexpr bool is_static   = false;
		static constexpr bool is_fixed    = false;
		static constexpr bool is_const    = true;

		constexpr const_storage_view()
			: m_data( nullptr ), m_size( 0 ) {}
		constexpr const_storage_view( const value_type* a_data, size_type a_size )
			: m_data( a_data ), m_size( a_size ) {}
		constexpr const_storage_view( const storage_view<T>& view )
			: m_data( view.data() ), m_size( view.size() ) {}

		void assign( const value_type* a_data, size_type a_size )
		{
			m_data = a_data;
			m_size = a_size;
		}

		constexpr size_type size() const { return m_size; }
		constexpr bool empty() const { return m_size != 0; }
		constexpr const value_type* data() const { return m_data; }
		constexpr size_type   raw_size() const { return sizeof( value_type ) * m_size; }
		constexpr const char* raw_data() const { return reinterpret_cast<const char*>( m_data ); }
	protected:
		const value_type* m_data;
		size_type         m_size;
	};


	template < typename T >
	inline void raw_construct_object( void* object )
	{
		new ( object )T;
	}

	template < typename T, typename ...Args >
	inline void raw_construct_object( void* object, Args&&... params )
	{
		new ( object )T( forward<Args>( params )... );
	}

	template < typename T >
	inline void raw_move_construct_object( void* object, T&& original )
	{
		new ( object )T( move( original ) );
	}

	template < typename T >
	inline void raw_copy_construct_object( void* object, const T& original )
	{
		new ( object )T( original );
	}

	template < typename T, typename ...Args >
	inline void construct_object( T* object, Args&&... params )
	{
		new ( object )T( forward<Args>( params )... );
	}

	template < typename T >
	inline void move_construct_object( T* object, T&& original )
	{
		new ( object )T( move( original ) );
	}

	template < typename T >
	inline void copy_construct_object( T* object, const T& original )
	{
		new ( object )T( original );
	}

	template < typename T >
	inline void destroy_object( T* object )
	{
		object->~T();
	}

	template <typename TYPE>
	void raw_destroy_object( void* object )
	{
		( (TYPE*)object )->TYPE::~TYPE();
	}

	namespace detail
	{
		template < typename ForwardIterator, typename T >
		void uninitialized_fill_impl( ForwardIterator first, ForwardIterator last, const T& value, true_type )
		{
			fill( first, last, value );
		}

		template < typename ForwardIterator, typename T >
		void uninitialized_fill_impl( ForwardIterator first, ForwardIterator last, const T& value, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; it != last; ++it )
				::new( (void*)&*it ) value_type( value );
		}

		template < typename ForwardIterator, typename T >
		ForwardIterator uninitialized_fill_n_impl( ForwardIterator first, size_t count, const T& value, true_type )
		{
			return fill_n( first, count, value );
		}

		template < typename ForwardIterator, typename T >
		ForwardIterator uninitialized_fill_n_impl( ForwardIterator first, size_t count, const T& value, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; count > 0; --count, ++it )
				::new ( (void*)&*it ) value_type( value );
			return it;
		}

		template < typename ForwardIterator >
		inline void uninitialized_construct_impl( ForwardIterator first, ForwardIterator last, true_type )
		{
			fill_default( first, last );
		}

		template < typename ForwardIterator >
		inline void uninitialized_construct_impl( ForwardIterator first, ForwardIterator last, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; it != last; ++it )
				::new( (void*)&*it ) value_type;
		}

		template < typename ForwardIterator >
		inline ForwardIterator uninitialized_construct_n_impl( ForwardIterator first, size_t count, true_type )
		{
			return fill_default_n( first, count );
		}

		template < typename ForwardIterator >
		inline ForwardIterator uninitialized_construct_n_impl( ForwardIterator first, size_t count, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; count > 0; --count, ++it )
				::new( (void*)&*it ) value_type;
		}

		template < typename ForwardIterator >
		inline void uninitialized_destroy_impl( ForwardIterator, ForwardIterator, true_type )
		{
			// no-op
		}

		template < typename ForwardIterator >
		inline void uninitialized_destroy_impl( ForwardIterator first, ForwardIterator last, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; it != last; ++it )
				( *it ).~value_type();
		}

		template < typename ForwardIterator >
		inline ForwardIterator uninitialized_destroy_n_impl( ForwardIterator first, size_t count, true_type )
		{
			return first + count;
		}

		template < typename ForwardIterator >
		inline ForwardIterator uninitialized_destroy_n_impl( ForwardIterator first, size_t count, false_type )
		{
			typedef typename iterator_traits< ForwardIterator >::value_type value_type;
			ForwardIterator it( first );
			for ( ; count > 0; --count, ++it )
				( *it ).~value_type();
			return it;
		}

		template < typename InputIterator, typename ForwardIterator >
		inline ForwardIterator uninitialized_copy_impl( InputIterator first, InputIterator last, ForwardIterator out, false_type )
		{
			typedef typename iterator_traits<ForwardIterator>::value_type value_type;
			InputIterator src( first );
			ForwardIterator dest( out );
			for ( ; src != last; ++src, ++dest )
			{
				::new ( static_cast<void*>( addressof( *dest ) ) ) value_type( *src );
			}
			return dest;
		}

		template < typename InputIterator, typename ForwardIterator >
		inline ForwardIterator uninitialized_copy_impl( InputIterator first, InputIterator last, ForwardIterator out, true_type )
		{
			return ForwardIterator( nvmemmove( out, first, (size_t)( (uintptr_t)last - (uintptr_t)first ) ) );
		}

		template < typename InputIterator, typename ForwardIterator >
		inline ForwardIterator uninitialized_copy_n_impl( InputIterator first, size_t count, ForwardIterator out, false_type )
		{
			typedef typename iterator_traits<ForwardIterator>::value_type value_type;
			InputIterator src( first );
			ForwardIterator dest( out );
			for ( ; count > 0; --count, ++src, ++dest )
			{
				::new ( static_cast<void*>( addressof( *dest ) ) ) value_type( *src );
			}
			return it;
		}

		template < typename InputIterator, typename ForwardIterator >
		inline ForwardIterator uninitialized_copy_n_impl( InputIterator first, size_t count, ForwardIterator out, true_type )
		{
			typedef typename iterator_traits<ForwardIterator>::value_type value_type;
			return ForwardIterator( nvmemmove( out, first, count * sizeof( value_type ) ) );
		}

	}

	template < typename ForwardIterator, typename T >
	inline void uninitialized_fill( ForwardIterator first, ForwardIterator last, const T& value )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		detail::uninitialized_fill_impl( first, last, value, has_trivial_destructor<value_type>() );
	}

	template < typename ForwardIterator, typename T >
	inline void uninitialized_fill_n( ForwardIterator first, size_t count, const T& value )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		return detail::uninitialized_fill_n_impl( first, count, value, has_trivial_destructor<value_type>() );
	}

	template < typename ForwardIterator, typename ...Args >
	inline void uninitialized_construct( ForwardIterator first, ForwardIterator last, Args&&... params )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		ForwardIterator it( first );
		for ( ; it != last; ++it )
			::new( (void*)&*it ) value_type( forward<Args>( params )... );
	}

	template < typename ForwardIterator >
	inline void uninitialized_construct( ForwardIterator first, ForwardIterator last )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		detail::uninitialized_construct_impl( first, last, has_trivial_destructor<value_type>() );
	}

	template < typename ForwardIterator >
	inline ForwardIterator uninitialized_construct_n( ForwardIterator first, size_t count )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		return detail::uninitialized_construct_n_impl( first, count, has_trivial_destructor<value_type>() );
	}

	template < typename ForwardIterator >
	inline void uninitialized_destroy( ForwardIterator first, ForwardIterator last )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		detail::uninitialized_destroy_impl( first, last, has_trivial_destructor<value_type>() );
	}

	template < typename ForwardIterator >
	inline ForwardIterator uninitialized_destroy_n( ForwardIterator first, size_t count )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		return detail::uninitialized_destroy_n_impl( first, count, has_trivial_destructor<value_type>() );
	}

	template < typename InputIterator, typename ForwardIterator >
	inline ForwardIterator uninitialized_copy( InputIterator first, InputIterator last, ForwardIterator out )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		return detail::uninitialized_copy_impl( first, last, out, has_trivial_copy<value_type>() );
	}

	template < typename InputIterator, typename ForwardIterator >
	inline ForwardIterator uninitialized_copy_n( InputIterator first, size_t count, ForwardIterator out )
	{
		typedef typename iterator_traits< ForwardIterator >::value_type value_type;
		return detail::uninitialized_copy_n_impl( first, count, out, has_trivial_copy<value_type>() );
	}

	namespace detail
	{

		template < typename Super, bool Const = Super::is_const >
		class add_iterators {};

		template < typename Super >
		class add_iterators < Super, true > : public Super
		{
		public:
			typedef typename Super::value_type           value_type;
			typedef const value_type*                    const_iterator;
			typedef nv::reverse_iterator<const_iterator> const_reverse_iterator;

			using Super::Super;

			inline const_iterator begin() const { return const_iterator( Super::data() ); }
			inline const_iterator end() const { return const_iterator( Super::data() + Super::size() ); }
			inline const_iterator cbegin() const { return const_iterator( Super::data() ); }
			inline const_iterator cend() const { return const_iterator( Super::data() + Super::size() ); }
			inline const_reverse_iterator rbegin() const { return const_reverse_iterator( end() ); }
			inline const_reverse_iterator crbegin() const { return const_reverse_iterator( end() ); }
			inline const_reverse_iterator rend() const { return const_reverse_iterator( begin() ); }
			inline const_reverse_iterator crend() const { return const_reverse_iterator( begin() ); }
			inline const_iterator iat( size_t i ) const { NV_ASSERT( i <= Super::size(), "Index out of range" ); return begin() + i; }
		};

		template < typename Super >
		class add_iterators < Super, false > : public Super
		{
		public:
			typedef typename Super::value_type           value_type;
			typedef value_type*                          iterator;
			typedef const value_type*                    const_iterator;
			typedef nv::reverse_iterator<iterator>       reverse_iterator;
			typedef nv::reverse_iterator<const_iterator> const_reverse_iterator;

			using Super::Super;

			inline const_iterator begin() const { return const_iterator( Super::data() ); }
			inline const_iterator end() const { return const_iterator( Super::data() + Super::size() ); }
			inline iterator begin() { return iterator( Super::data() ); }
			inline iterator end() { return iterator( Super::data() + Super::size() ); }
			inline const_iterator cbegin() const { return const_iterator( Super::data() ); }
			inline const_iterator cend() const { return const_iterator( Super::data() + Super::size() ); }
			inline reverse_iterator rbegin() { return reverse_iterator( end() ); }
			inline const_reverse_iterator rbegin() const { return const_reverse_iterator( end() ); }
			inline const_reverse_iterator crbegin() const { return const_reverse_iterator( end() ); }
			inline reverse_iterator rend() { return reverse_iterator( begin() ); }
			inline const_reverse_iterator rend() const { return const_reverse_iterator( begin() ); }
			inline const_reverse_iterator crend() const { return const_reverse_iterator( begin() ); }
			inline const_iterator iat( size_t i ) const { NV_ASSERT( i <= Super::size(), "Index out of range" ); return begin() + i; }
			inline iterator       iat( size_t i ) { NV_ASSERT( i <= Super::size(), "Index out of range" ); return begin() + i; }
		};

		template < typename Super, bool Const = Super::is_const >
		class add_random_access {};

		template < typename Super >
		class add_random_access < Super, true > : public Super
		{
		public:
			typedef typename Super::value_type           value_type;
			typedef typename Super::size_type            size_type;
			typedef const value_type*                    const_iterator;
			typedef value_type&                          reference;
			typedef const value_type&                    const_reference;

			using Super::Super;

			inline const_reference front() const { NV_ASSERT( !Super::empty(), "front() called on empty data!" ); return Super::data()[0]; }
			inline const_reference back() const { NV_ASSERT( !Super::empty(), "front() called on empty data!" ); return Super::data()[Super::size() - 1]; }

			const_reference operator[]( size_type i ) const
			{
				NV_ASSERT( i < Super::size(), "Out of range" );
				return Super::data()[i];
			}
		};

		template < typename Super >
		class add_random_access < Super, false > : public Super
		{
		public:
			typedef typename Super::value_type           value_type;
			typedef typename Super::size_type            size_type;
			typedef value_type*                          iterator;
			typedef const value_type*                    const_iterator;
			typedef value_type&                          reference;
			typedef const value_type&                    const_reference;

			using Super::Super;

			inline reference       front() { NV_ASSERT( !Super::empty(), "front() called on empty data!" );  return Super::data()[0]; }
			inline const_reference front() const { NV_ASSERT( !Super::empty(), "front() called on empty data!" ); return Super::data()[0]; }
			inline reference       back() { NV_ASSERT( !Super::empty(), "front() called on empty data!" ); return Super::data()[Super::size() - 1]; }
			inline const_reference back() const { NV_ASSERT( !Super::empty(), "front() called on empty data!" ); return Super::data()[Super::size() - 1]; }

			reference operator[]( size_type i )
			{
				NV_ASSERT( i < Super::size(), "Out of range" );
				return Super::data()[i];
			}

			const_reference operator[]( size_type i ) const
			{
				NV_ASSERT( i < Super::size(), "Out of range" );
				return Super::data()[i];
			}

			inline void fill( const value_type& value )
			{
				fill_n( iterator( Super::data() ), iterator( Super::data() + this->size() ), value );
			}

		};

	}

	template < typename T > 
	using const_array_ref = detail::add_random_access< detail::add_iterators < const_storage_view< T > > >;
	template < typename T >
	using array_ref       = detail::add_random_access< detail::add_iterators < storage_view< T > > >;
	using const_mem_ref = const_array_ref< char >;
	using mem_ref       = array_ref< char >;

}

#endif // NV_CORE_MEMORY_HH
